#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Программа для захвата видео с камеры ZWO ASI, управления параметрами
и помощи в полярном выравнивании по трем звездам.
Отслеживание звезд по центроиду яркости.
Основано на Tkinter GUI.
"""

import numpy as np
import tkinter as tk
from tkinter import ttk, messagebox
import threading
import time
import cv2
from PIL import Image, ImageTk, ImageFont, ImageDraw 
import traceback 
import math
import os 

try:
    import zwoasi as asi
except ImportError:
    messagebox.showerror("Ошибка", "Модуль zwoasi не найден.\nУстановите: pip install zwoasi")
    exit(1)

class CameraTrackerApp:
    def __init__(self):
        self.root = tk.Tk()
        self.root.title("ZWO ASI Polar Alignment Tool v7") 
        self.root.geometry("1000x750")

        self.camera = None
        self.is_capturing = False
        self.capture_thread = None
        self.camera_info = {}
        self.all_control_caps = {}
        self.available_cameras_info = []
        self.current_raw_frame = None 

        self.exposure_us = 20000  
        self.gain = 150 
        self.usb_traffic = 50 

        self.lock_on_roi_size = (30, 30) 
        self.search_window_size = (80, 80) 
        self.star_brightness_threshold = 60 
        
        self.polar_align_mode = False 
        self.polar_stars_data = {}  
        self.polar_star_names_ordered = ['polaris', 'hip7283', 'hip3128'] 
        self.polar_star_display_names = { 
            'polaris': 'Полярная (Polaris)',
            'hip7283': 'HIP 7283',
            'hip3128': 'HIP 3128'
        }
        self.polar_star_sky_coords_deg = {
            'polaris': {'ra': (2 + 30/60 + 21.61/3600) * 15, 'dec': 89 + 15/60 + 41.9/3600},
            'hip7283': {'ra': (1 + 33/60 + 1.29/3600) * 15, 'dec': 89 + 0/60 + 41.9/3600},
            'hip3128': {'ra': (0 + 38/60 + 26.77/3600) * 15, 'dec': 89 + 26/60 + 22.8/3600}
        }
        self.current_polar_star_to_select_idx = 0 
        self.polar_pole_calculated_coords = None 
        self.show_polar_grid = False 
        self.pixels_per_degree_estimate = 100.0 
        self.show_center_crosshair = False 
        
        self.font_size_status = 18
        self.font_size_overlay = 14
        self.pil_font_status = None
        self.pil_font_overlay = None
        try:
            font_path = "arial.ttf" 
            if os.name != 'nt': 
                 possible_linux_font_paths = ["/usr/share/fonts/truetype/dejavu/DejaVuSans.ttf", "DejaVuSans.ttf"]
                 found_font = False
                 for p in possible_linux_font_paths:
                     try:
                         self.pil_font_status = ImageFont.truetype(p, self.font_size_status)
                         self.pil_font_overlay = ImageFont.truetype(p, self.font_size_overlay)
                         print(f"Загружен шрифт: {p}"); found_font = True; break
                     except IOError: continue
                 if not found_font: print("Шрифт Arial/DejaVuSans не найден."); self.pil_font_status=ImageFont.load_default(); self.pil_font_overlay=ImageFont.load_default()
            else: 
                 self.pil_font_status = ImageFont.truetype(font_path, self.font_size_status)
                 self.pil_font_overlay = ImageFont.truetype(font_path, self.font_size_overlay)
                 print(f"Загружен шрифт: {font_path}")
        except IOError:
            print(f"Ошибка: Шрифт '{font_path}' не найден."); self.pil_font_status=ImageFont.load_default(); self.pil_font_overlay=ImageFont.load_default()
                                           
        self.setup_gui()

        try:
            asi.init(); self.status_var.set("ZWO ASI SDK инициализирован. Найдите камеры."); print("ZWO ASI SDK инициализирован")
        except Exception as e: messagebox.showerror("Ошибка SDK", f"Не удалось инициализировать ZWO ASI SDK: {e}"); self.status_var.set(f"Ошибка SDK: {e}")

    def draw_text_pil(self, cv_image_bgr, text, position, font, bgr_color_for_pil):
        try:
            rgb_image = cv2.cvtColor(cv_image_bgr, cv2.COLOR_BGR2RGB)
            pil_image = Image.fromarray(rgb_image)
            draw = ImageDraw.Draw(pil_image)
            # PIL ожидает RGB, поэтому если пользователь указал (255,0,0) для красного, это уже RGB
            # bgr_color_for_pil должен быть уже в формате (R,G,B) если он пришел от пользователя
            # или (B,G,R) если это цвет для cv2 функций.
            # Для простоты, эта функция теперь ожидает цвет в BGR и сама конвертирует.
            rgb_color_pil = (bgr_color_for_pil[2], bgr_color_for_pil[1], bgr_color_for_pil[0])
            draw.text(position, text, font=font, fill=rgb_color_pil)
            return cv2.cvtColor(np.array(pil_image), cv2.COLOR_RGB2BGR)
        except Exception as e:
            print(f"Ошибка PIL text: {e}"); cv2.putText(cv_image_bgr, "TXT_ERR", position, cv2.FONT_HERSHEY_SIMPLEX, 0.5, (0,0,255), 1); return cv_image_bgr

    def setup_gui(self):
        main_frame = ttk.Frame(self.root, padding="10"); main_frame.pack(fill=tk.BOTH, expand=True)
        left_panel = ttk.Frame(main_frame, width=280); left_panel.pack(side=tk.LEFT, fill=tk.Y, padx=(0, 10)); left_panel.pack_propagate(False)
        right_panel = ttk.Frame(main_frame); right_panel.pack(side=tk.RIGHT, fill=tk.BOTH, expand=True)
        self.setup_control_panel(left_panel); self.setup_video_panel(right_panel)
        self.root.protocol("WM_DELETE_WINDOW", self.on_closing)

    def setup_control_panel(self, parent):
        camera_frame = ttk.LabelFrame(parent, text="Управление камерой"); camera_frame.pack(fill=tk.X, pady=(0, 10))
        self.find_btn = ttk.Button(camera_frame, text="Найти камеры", command=self.find_cameras); self.find_btn.pack(fill=tk.X, padx=5, pady=5)
        self.camera_listbox = tk.Listbox(camera_frame, height=3, exportselection=False); self.camera_listbox.pack(fill=tk.X, padx=5, pady=(0,5))
        self.connect_btn = ttk.Button(camera_frame, text="Подключить", command=self.connect_camera, state=tk.DISABLED); self.connect_btn.pack(fill=tk.X, padx=5, pady=(0,5))
        self.start_btn = ttk.Button(camera_frame, text="Запустить видео", command=self.start_capture, state=tk.DISABLED); self.start_btn.pack(fill=tk.X, padx=5, pady=(0,5))
        self.stop_btn = ttk.Button(camera_frame, text="Остановить видео", command=self.stop_capture, state=tk.DISABLED); self.stop_btn.pack(fill=tk.X, padx=5, pady=5)

        settings_frame = ttk.LabelFrame(parent, text="Настройки камеры"); settings_frame.pack(fill=tk.X, pady=(0, 10))
        ttk.Label(settings_frame, text="Экспозиция (мс):").pack(anchor=tk.W, padx=5, pady=(5,0))
        self.exposure_var = tk.DoubleVar(value=self.exposure_us / 1000.0)
        self.exposure_scale = ttk.Scale(settings_frame, from_=1, to=5000, variable=self.exposure_var, orient=tk.HORIZONTAL, command=self.on_exposure_change); self.exposure_scale.pack(fill=tk.X, padx=5)
        self.exposure_label_var = tk.StringVar(value=f"{self.exposure_us / 1000.0:.1f} мс"); ttk.Label(settings_frame, textvariable=self.exposure_label_var).pack(anchor=tk.W, padx=5)
        ttk.Label(settings_frame, text="Усиление:").pack(anchor=tk.W, padx=5, pady=(5,0))
        self.gain_var = tk.IntVar(value=self.gain)
        self.gain_scale = ttk.Scale(settings_frame, from_=0, to=100, variable=self.gain_var, orient=tk.HORIZONTAL, command=self.on_gain_change); self.gain_scale.pack(fill=tk.X, padx=5)
        self.gain_label_var = tk.StringVar(value=f"{self.gain}"); ttk.Label(settings_frame, textvariable=self.gain_label_var).pack(anchor=tk.W, padx=5, pady=(0,5))
        ttk.Label(settings_frame, text="USB Bandwidth Val (SDK):").pack(anchor=tk.W, padx=5, pady=(5,0))
        self.usb_traffic_var = tk.IntVar(value=self.usb_traffic)
        self.usb_traffic_scale = ttk.Scale(settings_frame, from_=0, to=100, variable=self.usb_traffic_var, orient=tk.HORIZONTAL, command=self.on_usb_traffic_change); self.usb_traffic_scale.pack(fill=tk.X, padx=5)
        self.usb_traffic_label_var = tk.StringVar(value=f"{self.usb_traffic}"); ttk.Label(settings_frame, textvariable=self.usb_traffic_label_var).pack(anchor=tk.W, padx=5, pady=(0,5))
        
        polar_align_frame = ttk.LabelFrame(parent, text="Полярное выравнивание"); polar_align_frame.pack(fill=tk.X, pady=(10,10))
        self.start_polar_align_btn = ttk.Button(polar_align_frame, text="Начать/Сбросить калибровку", command=self.toggle_polar_alignment_mode, state=tk.DISABLED); self.start_polar_align_btn.pack(fill=tk.X, padx=5, pady=5)
        self.toggle_grid_btn = ttk.Button(polar_align_frame, text="Показать/Скрыть экв. сетку", command=self.toggle_polar_grid_display, state=tk.DISABLED); self.toggle_grid_btn.pack(fill=tk.X, padx=5, pady=5)
        self.center_crosshair_btn = ttk.Button(polar_align_frame, text="Вкл/Выкл центр. перекрестие", command=self.toggle_center_crosshair, state=tk.DISABLED); self.center_crosshair_btn.pack(fill=tk.X, padx=5, pady=5)
        
        self.status_var = tk.StringVar(value="Ожидание действий...")
        status_label = ttk.Label(parent, textvariable=self.status_var, wraplength=260, relief=tk.SUNKEN, anchor=tk.W); status_label.pack(fill=tk.X, pady=10, ipady=5)
    
    def setup_video_panel(self, parent):
        video_frame = ttk.LabelFrame(parent, text="Видеопоток"); video_frame.pack(fill=tk.BOTH, expand=True)
        self.video_canvas = tk.Canvas(video_frame, bg='black'); self.video_canvas.pack(fill=tk.BOTH, expand=True)
        self.video_canvas.bind('<Button-1>', self.on_canvas_click)
        self.video_canvas.create_text(350, 250, text="Камера не подключена", fill='white', font=('Arial', 14), tags="infotext")
        self.canvas_img_scale = 1.0; self.canvas_img_offset_x = 0; self.canvas_img_offset_y = 0

    def find_cameras(self):
        self.status_var.set("Поиск камер..."); self.root.update_idletasks()
        try:
            num_cameras = asi.get_num_cameras(); self.camera_listbox.delete(0, tk.END); self.available_cameras_info = [] 
            if num_cameras == 0: self.status_var.set("Камеры не найдены."); messagebox.showwarning("Поиск", "Камеры не найдены."); self.connect_btn.config(state=tk.DISABLED); return
            names = []; 
            try: names = asi.list_cameras()
            except: print("WARN: asi.list_cameras() fail")
            for i in range(num_cameras):
                name = names[i] if i < len(names) and names[i] else f"ASI Camera #{i}" 
                self.available_cameras_info.append({'id': i, 'name': name}); self.camera_listbox.insert(tk.END, f"{i}: {name}")
            self.status_var.set(f"Найдено: {num_cameras}. Выберите."); self.connect_btn.config(state=tk.NORMAL)
            if num_cameras > 0: self.camera_listbox.selection_set(0)
        except Exception as e: messagebox.showerror("Ошибка", f"Ошибка поиска: {e}"); self.status_var.set(f"Ошибка: {e}"); self.connect_btn.config(state=tk.DISABLED)

    def connect_camera(self):
        sel = self.camera_listbox.curselection()
        if not sel: messagebox.showwarning("Подключение", "Выберите камеру."); return
        idx = sel[0]; name = self.available_cameras_info[idx]['name'] if idx < len(self.available_cameras_info) else "Неизвестно"
        self.status_var.set(f"Подключение: {name}..."); self.root.update_idletasks()
        try:
            if self.camera: self.stop_capture(disconnect_camera=True) 
            self.camera = asi.Camera(idx) 
            try: self.camera_info = self.camera.get_camera_property()
            except: print("WARN: get_camera_property() fail"); self.camera_info = {'Name': name, 'MaxWidth': 1280, 'MaxHeight': 960, 'SupportedVideoFormat': []}
            try: self.all_control_caps = self.camera.get_control_caps()
            except: print("WARN: get_control_caps() fail"); self.all_control_caps = {} 
            actual_name = self.camera_info.get('Name', name); print(f"Подключена: {actual_name}")
            formats = self.camera_info.get('SupportedVideoFormat', []); fmt_to_set = None
            img_type_enum = getattr(asi, 'ASI_IMG_TYPE', asi)
            possible_formats = []
            for attr_name in ['ASI_IMG_RAW8', 'ASI_IMG_MONO8', 'ASI_IMG_RAW16']:
                if hasattr(img_type_enum, attr_name): possible_formats.append(getattr(img_type_enum, attr_name))
            for p_fmt in possible_formats:
                if p_fmt in formats: fmt_to_set = p_fmt; break
            if fmt_to_set: self.camera.set_image_type(fmt_to_set); print(f"Тип изображения: {fmt_to_set}")
            else: print(f"WARN: RAW/MONO не найден. Тип: {self.camera.get_image_type()}")
            self.camera.set_roi_format( self.camera_info.get('MaxWidth',1280), self.camera_info.get('MaxHeight',960),1,self.camera.get_image_type())
            print("ROI установлен.")
            self.update_gui_controls_from_camera(); self.apply_current_settings_to_camera() 
            self.status_var.set(f"Подключена: {actual_name}")
            self.connect_btn.config(state=tk.DISABLED); self.start_btn.config(state=tk.NORMAL)     
            self.stop_btn.config(state=tk.DISABLED); self.start_polar_align_btn.config(state=tk.DISABLED); self.toggle_grid_btn.config(state=tk.DISABLED); self.center_crosshair_btn.config(state=tk.DISABLED)
        except Exception as e: 
            traceback.print_exc(); messagebox.showerror("Ошибка", f"Не удалось подключиться: {e}"); self.status_var.set(f"Ошибка: {e}")
            if self.camera: _ = getattr(self.camera, 'close', lambda: None)(); self.camera = None 
            self.start_btn.config(state=tk.DISABLED); self.connect_btn.config(state=tk.NORMAL)

    def update_gui_controls_from_camera(self):
        if not self.camera: return
        try: 
            val, _ = self.camera.get_control_value(asi.ASI_EXPOSURE); self.exposure_var.set(val/1000.0); self.exposure_label_var.set(f"{val/1000.0:.1f} мс"); self.exposure_us=val 
            s_min,s_max=1,5000
            if self.all_control_caps and asi.ASI_EXPOSURE in self.all_control_caps and isinstance(self.all_control_caps[asi.ASI_EXPOSURE],dict):
                cap=self.all_control_caps[asi.ASI_EXPOSURE]; s_min=max(1,cap.get('MinValue',1000)//1000); s_max=min(5000,cap.get('MaxValue',5000000)//1000)
            self.exposure_scale.config(from_=s_min, to=s_max)
        except: print("Ошибка GUI Exp"); self.exposure_scale.config(from_=1,to=5000)
        try:
            val, _ = self.camera.get_control_value(asi.ASI_GAIN); self.gain_var.set(val); self.gain_label_var.set(f"{val}"); self.gain=val
            s_min,s_max=0,100
            if self.all_control_caps and asi.ASI_GAIN in self.all_control_caps and isinstance(self.all_control_caps[asi.ASI_GAIN],dict):
                cap=self.all_control_caps[asi.ASI_GAIN]; s_min=cap.get('MinValue',0); s_max=cap.get('MaxValue',100)
            self.gain_scale.config(from_=s_min, to=s_max)
        except: print("Ошибка GUI Gain"); self.gain_scale.config(from_=0,to=100)
        try:
            val = self.usb_traffic; s_min,s_max=40,100
            if self.all_control_caps and asi.ASI_BANDWIDTHOVERLOAD in self.all_control_caps and isinstance(self.all_control_caps[asi.ASI_BANDWIDTHOVERLOAD],dict):
                cap=self.all_control_caps[asi.ASI_BANDWIDTHOVERLOAD]; s_min=cap.get('MinValue',40); s_max=cap.get('MaxValue',100)
                try: val_cam,_=self.camera.get_control_value(asi.ASI_BANDWIDTHOVERLOAD); val=val_cam; self.usb_traffic=val_cam
                except: print("Не удалось прочитать USB BW")
            self.usb_traffic_var.set(val); self.usb_traffic_label_var.set(f"{val}"); self.usb_traffic_scale.config(from_=s_min, to=s_max)
        except: print("Ошибка GUI USB"); self.usb_traffic_scale.config(from_=40,to=100)

    def apply_current_settings_to_camera(self):
        if not self.camera: return
        try:
            print(f"Применение: Exp={int(self.exposure_us)}us, Gain={int(self.gain)}, USBVal={int(self.usb_traffic)}")
            self.camera.set_control_value(asi.ASI_EXPOSURE, int(self.exposure_us))
            self.camera.set_control_value(asi.ASI_GAIN, int(self.gain))
            bw_val = int(self.usb_traffic)
            if hasattr(asi, 'ASI_BANDWIDTHOVERLOAD'):
                s_min,s_max=40,100
                if self.all_control_caps and asi.ASI_BANDWIDTHOVERLOAD in self.all_control_caps and isinstance(self.all_control_caps[asi.ASI_BANDWIDTHOVERLOAD],dict):
                    cap=self.all_control_caps[asi.ASI_BANDWIDTHOVERLOAD]; s_min=cap.get('MinValue',40); s_max=cap.get('MaxValue',100)
                bw_val = max(s_min, min(s_max, bw_val))
                try: self.camera.set_control_value(asi.ASI_BANDWIDTHOVERLOAD, bw_val); print(f"USB BW установлен: {bw_val}")
                except asi.ZWO_Error as e: print(f"Ошибка уст. USB BW: {e}" if not hasattr(e,'code') or e.code!=8 else "Контрол USB BW не подд.")
            self.status_var.set(f"Exp:{self.exposure_us/1000:.1f}ms, Gain:{self.gain}, USB:{bw_val}")
        except Exception as e: print(f"Ошибка применения: {e}"); self.status_var.set(f"Ошибка: {e}")
    
    def on_exposure_change(self, v): self.exposure_us=int(float(v)*1000); self.exposure_label_var.set(f"{float(v):.1f} мс"); self._conditional_apply_settings()
    def on_gain_change(self, v): self.gain=int(float(v)); self.gain_label_var.set(f"{self.gain}"); self._conditional_apply_settings()
    def on_usb_traffic_change(self, v): self.usb_traffic=int(float(v)); self.usb_traffic_label_var.set(f"{self.usb_traffic}"); self._conditional_apply_settings()
    def _conditional_apply_settings(self):
        if self.camera and self.is_capturing: self.apply_current_settings_to_camera()

    def start_capture(self):
        if not self.camera: messagebox.showerror("Ошибка", "Камера не подключена."); return
        if self.is_capturing: messagebox.showwarning("Захват", "Захват уже активен."); return
        try:
            self.apply_current_settings_to_camera(); self.camera.start_video_capture(); self.is_capturing=True
            self.capture_thread=threading.Thread(target=self._capture_loop,daemon=True); self.capture_thread.start()
            self.start_btn.config(state=tk.DISABLED); self.stop_btn.config(state=tk.NORMAL)
            self.connect_btn.config(state=tk.DISABLED); self.start_polar_align_btn.config(state=tk.NORMAL)
            self.center_crosshair_btn.config(state=tk.NORMAL) 
            self.status_var.set("Видеопоток активен."); self.video_canvas.delete("infotext")
        except Exception as e: messagebox.showerror("Ошибка",f"Не удалось запустить: {e}"); self.is_capturing=False

    def stop_capture(self, disconnect_camera=False):
        cap=self.is_capturing; self.is_capturing=False
        if self.capture_thread and self.capture_thread.is_alive(): print("Ожидание потока...");self.capture_thread.join(timeout=1.0)
        if self.camera:
            try:
                if cap: self.camera.stop_video_capture();print("Видео остановлено.")
                if disconnect_camera: self.camera.close();self.camera=None;self.all_control_caps={};self.camera_info={};print("Камера откл.");self.connect_btn.config(state=tk.NORMAL)
            except Exception as e:print(f"Ошибка остановки: {e}")
        self.reset_polar_alignment_state(update_gui=False)
        self.start_btn.config(state=tk.NORMAL if self.camera else tk.DISABLED);self.stop_btn.config(state=tk.DISABLED)
        self.start_polar_align_btn.config(state=tk.DISABLED);self.toggle_grid_btn.config(state=tk.DISABLED)
        self.center_crosshair_btn.config(state=tk.DISABLED) 
        stat="Видео остановлено." if self.camera and not disconnect_camera else "Камера отключена."
        if self.camera and not disconnect_camera: stat=f"Видео остановлено. Камера: {self.camera_info.get('Name','N/A')}"
        self.status_var.set(stat)
        if disconnect_camera or not self.camera: self.video_canvas.delete("all");self.video_canvas.create_text(350,250,text="Камера не подключена",fill='white',font=('Arial',14),tags="infotext")

    def _prepare_frame_for_processing(self, frame):
        if frame is None: return None
        if frame.dtype==np.uint16: frame=(cv2.normalize(frame,None,0,255,cv2.NORM_MINMAX,dtype=cv2.CV_8U) if np.max(frame)>0 else frame.astype(np.uint8))
        elif len(frame.shape)==3: frame=cv2.cvtColor(frame,cv2.COLOR_BGR2GRAY)
        elif frame.dtype!=np.uint8: 
            try: frame=frame.astype(np.uint8)
            except: return None
        return frame

    def _find_brightest_centroid_in_roi(self, roi_gray, roi_x, roi_y):
        if roi_gray is None or roi_gray.size==0: return None
        _,max_val,_,max_loc = cv2.minMaxLoc(roi_gray)
        if max_val<self.star_brightness_threshold: return None
        sx,sy=max_loc; sr_size=15
        s_x1=max(0,sx-sr_size//2); s_y1=max(0,sy-sr_size//2)
        s_x2=min(roi_gray.shape[1],sx+sr_size//2); s_y2=min(roi_gray.shape[0],sy+sr_size//2)
        sub=roi_gray[s_y1:s_y2,s_x1:s_x2]
        if sub.size==0: return (sx+roi_x, sy+roi_y)
        _,thresh=cv2.threshold(sub,max_val*0.7,255,cv2.THRESH_BINARY)
        contours,_=cv2.findContours(thresh,cv2.RETR_EXTERNAL,cv2.CHAIN_APPROX_SIMPLE)
        if contours:
            cnt=max(contours,key=cv2.contourArea)
            if cv2.contourArea(cnt)>0:
                M=cv2.moments(cnt)
                if M["m00"]>1e-5: return (s_x1+int(M["m10"]/M["m00"])+roi_x, s_y1+int(M["m01"]/M["m00"])+roi_y)
        return (sx+roi_x, sy+roi_y)

    def on_canvas_click(self, event):
        if not self.is_capturing or self.current_raw_frame is None: self.status_var.set("Запустите видео."); return
        cx,cy=event.x,event.y; off_x,off_y=self.canvas_img_offset_x,self.canvas_img_offset_y; scale=self.canvas_img_scale
        orig_x=int((cx-off_x)/scale) if scale>1e-6 else 0; orig_y=int((cy-off_y)/scale) if scale>1e-6 else 0
        fh,fw=self.current_raw_frame.shape[:2]
        if not (0<=orig_x<fw and 0<=orig_y<fh): self.status_var.set("Клик вне кадра."); return
        frame=self._prepare_frame_for_processing(self.current_raw_frame)
        if frame is None: self.status_var.set("Ошибка подготовки кадра."); return

        if self.polar_align_mode:
            if self.current_polar_star_to_select_idx < len(self.polar_star_names_ordered):
                key=self.polar_star_names_ordered[self.current_polar_star_to_select_idx]; name=self.polar_star_display_names[key]
                self.status_var.set(f"Захват '{name}' в ({orig_x},{orig_y})...")
                rx1=max(0,orig_x-self.lock_on_roi_size[0]//2); ry1=max(0,orig_y-self.lock_on_roi_size[1]//2)
                rx2=min(fw,orig_x+self.lock_on_roi_size[0]//2); ry2=min(fh,orig_y+self.lock_on_roi_size[1]//2)
                roi=frame[ry1:ry2,rx1:rx2]
                if roi.size==0: self.status_var.set("ROI захвата 0."); return
                coords=self._find_brightest_centroid_in_roi(roi,rx1,ry1)
                if coords:
                    self.polar_stars_data[key]={'screen_coords':coords,'sky_coords':self.polar_star_sky_coords_deg[key],'lost_tracking':False}
                    print(f"'{name}' захвачена: {coords}"); self.current_polar_star_to_select_idx+=1
                    if self.current_polar_star_to_select_idx<len(self.polar_star_names_ordered):
                        nkey=self.polar_star_names_ordered[self.current_polar_star_to_select_idx]
                        self.status_var.set(f"Отлично! Кликните на '{self.polar_star_display_names[nkey]}'.")
                    else:
                        self.status_var.set("3 звезды выбраны! Расчет полюса..."); self.calculate_and_set_pole_position()
                        self.toggle_grid_btn.config(state=tk.NORMAL if self.polar_pole_calculated_coords else tk.DISABLED)
                else: self.status_var.set(f"Не удалось захватить '{name}'.")
            else: self.status_var.set("Все звезды выбраны. Сбросьте для выбора заново.")
        else: self.status_var.set("Для выбора звезд активируйте 'Начать калибровку'.")

    def _capture_loop(self):
        last_fps_t=time.time(); fc=0; fps=0
        while self.is_capturing and self.camera:
            try:
                raw=self.camera.capture_video_frame(timeout=max(self.exposure_us//1000*2+200,500))
                if raw is None: time.sleep(0.01); continue
                frame=self._prepare_frame_for_processing(raw)
                if frame is None: continue
                self.current_raw_frame=frame.copy()
                
                all_tracked=True; num_sel=0
                if self.polar_align_mode or self.show_polar_grid:
                    for key,data in self.polar_stars_data.items():
                        if 'screen_coords' in data and data['screen_coords']:
                            num_sel+=1
                            if not data.get('lost_tracking',False):
                                px,py=data['screen_coords']
                                sx1=max(0,px-self.search_window_size[0]//2);sy1=max(0,py-self.search_window_size[1]//2)
                                sx2=min(frame.shape[1],px+self.search_window_size[0]//2);sy2=min(frame.shape[0],py+self.search_window_size[1]//2)
                                sroi=frame[sy1:sy2,sx1:sx2]
                                if sroi.size>0:
                                    found=self._find_brightest_centroid_in_roi(sroi,sx1,sy1)
                                    if found: self.polar_stars_data[key]['screen_coords']=found
                                    else: data['lost_tracking']=True; all_tracked=False
                                else: data['lost_tracking']=True; all_tracked=False
                            else: all_tracked=False
                        else: all_tracked=False
                    if num_sel==3 and all_tracked and (self.polar_align_mode or self.show_polar_grid): 
                        self.calculate_and_set_pole_position(silent=True) 
                
                bgr=cv2.cvtColor(frame,cv2.COLOR_GRAY2BGR)
                if self.show_center_crosshair:
                    h,w=bgr.shape[:2];cx,cy=w//2,h//2; color=(255,255,255) # Белый
                    cv2.line(bgr,(0,cy),(w,cy),color,1); cv2.line(bgr,(cx,0),(cx,h),color,1)

                if self.polar_align_mode or self.show_polar_grid: 
                    for star_key, data in self.polar_stars_data.items():
                        if 'screen_coords' in data and data['screen_coords'] is not None:
                            sx, sy = map(int, data['screen_coords']) 
                            marker_color_bgr = (0,0,255) if data.get('lost_tracking') else (0,255,0) # Красный если потерян, зеленый если отслеживается
                            cv2.line(bgr, (sx - 7, sy), (sx + 7, sy), marker_color_bgr, 1) # Горизонтальная часть креста
                            cv2.line(bgr, (sx, sy - 7), (sx, sy + 7), marker_color_bgr, 1) # Вертикальная
                            text_to_draw = self.polar_star_display_names.get(star_key, star_key)
                            bgr = self.draw_text_pil(bgr, text_to_draw, (sx + 10, sy - 20), self.pil_font_overlay, marker_color_bgr)
                            if data.get('lost_tracking'):
                                bgr = self.draw_text_pil(bgr, "Lost", (sx + 10, sy + 0), self.pil_font_overlay, (0,0,255)) # Красный

                    if self.show_polar_grid and self.polar_pole_calculated_coords:
                        # Передаем bgr, чтобы функция могла на нем рисовать и возвращать измененный
                        bgr = self.draw_polar_grid_on_frame(bgr, self.polar_pole_calculated_coords)
                
                if self.polar_align_mode and self.current_polar_star_to_select_idx < len(self.polar_star_names_ordered):
                    star_key_to_select = self.polar_star_names_ordered[self.current_polar_star_to_select_idx]
                    status_text = f"Select: {self.polar_star_display_names[star_key_to_select]}"
                    bgr = self.draw_text_pil(bgr, status_text, (10, bgr.shape[0] - 30), self.pil_font_status, (0,255,255)) # Желтый для подсказки (BGR)

                fc+=1; ct=time.time()
                if ct-last_fps_t>=1.0: fps=fc/(ct-last_fps_t);fc=0;last_fps_t=ct
                fps_text = f"FPS: {fps:.1f}"
                bgr = self.draw_text_pil(bgr, fps_text, (10,10), self.pil_font_status, (0,255,0)) # Зеленый для FPS (BGR)
                
                if hasattr(self,'root') and self.root.winfo_exists(): self.root.after(0,self.update_canvas_display,bgr)
            except asi.ZWO_Error as e:
                if hasattr(e,'code') and e.code==asi.ASI_ERROR_TIMEOUT: time.sleep(0.01); continue
                print(f"ASI ошибка: {e}");self.is_capturing=False; self.root.after(0,lambda:self.status_var.set(f"ASI Ошибка: {e}")); break
            except Exception as e: print(f"Ошибка в цикле: {e}");traceback.print_exc();self.is_capturing=False; self.root.after(0,lambda:self.status_var.set(f"Ошибка: {e}")); break
        print("Цикл захвата завершен.")
        if hasattr(self,'root') and self.root.winfo_exists(): self.root.after(0,lambda:self.status_var.set("Видео остановлено."))

    def update_canvas_display(self, bgr_frame):
        if not hasattr(self,'video_canvas') or not self.video_canvas.winfo_exists(): return
        if bgr_frame is None: return
        try:
            ih,iw=bgr_frame.shape[:2]; cw=self.video_canvas.winfo_width(); ch=self.video_canvas.winfo_height()
            if cw<=1 or ch<=1: return
            sw=cw/iw; sh=ch/ih; self.canvas_img_scale=min(sw,sh)
            dw=int(iw*self.canvas_img_scale); dh=int(ih*self.canvas_img_scale)
            if dw<=0 or dh<=0: return
            # PIL ожидает RGB. cv2.cvtColor(frame, cv2.COLOR_GRAY2BGR) уже создал BGR.
            # Для корректного отображения цветов через PIL, конвертируем BGR в RGB
            img_rgb = cv2.cvtColor(bgr_frame, cv2.COLOR_BGR2RGB)
            resized_img = cv2.resize(img_rgb,(dw,dh),interpolation=cv2.INTER_LINEAR) 
            pil=Image.fromarray(resized_img); self.tk_photo=ImageTk.PhotoImage(image=pil)
            self.canvas_img_offset_x=(cw-dw)//2; self.canvas_img_offset_y=(ch-dh)//2
            self.video_canvas.delete("all"); self.video_canvas.create_image(self.canvas_img_offset_x,self.canvas_img_offset_y,anchor=tk.NW,image=self.tk_photo)
        except Exception as e:
            if hasattr(self,'video_canvas') and self.video_canvas.winfo_exists(): print(f"Ошибка Canvas: {e}")
    
    def toggle_polar_alignment_mode(self):
        if not self.is_capturing: messagebox.showwarning("Полярка", "Сначала запустите видеопоток."); return
        if self.polar_align_mode: self.reset_polar_alignment_state(); self.status_var.set("Калибровка полярки сброшена. Кликните для начала.")
        else: 
            self.reset_polar_alignment_state(); self.polar_align_mode = True
            star_key = self.polar_star_names_ordered[self.current_polar_star_to_select_idx]
            self.status_var.set(f"Режим полярки: Кликните на '{self.polar_star_display_names[star_key]}'.")
            print("Режим калибровки полярки активирован.")
    
    def reset_polar_alignment_state(self, update_gui=True):
        self.polar_align_mode = False; self.polar_stars_data = {}; self.current_polar_star_to_select_idx = 0
        self.polar_pole_calculated_coords = None; self.show_polar_grid = False; self.affine_transform_matrix = None
        if update_gui and hasattr(self, 'toggle_grid_btn'): 
            self.toggle_grid_btn.config(state=tk.DISABLED, text="Показать сетку")
            if hasattr(self, 'status_var'): self.status_var.set("Калибровка полярки сброшена.")
        print("Состояние полярной калибровки сброшено.")

    def gnomonic_projection(self, ra_deg, dec_deg, ra0_deg, dec0_deg):
        ra_rad, dec_rad = math.radians(ra_deg), math.radians(dec_deg)
        ra0_rad, dec0_rad = math.radians(ra0_deg), math.radians(dec0_deg)
        cos_c = math.sin(dec0_rad) * math.sin(dec_rad) + \
                math.cos(dec0_rad) * math.cos(dec_rad) * math.cos(ra_rad - ra0_rad)
        if abs(cos_c) < 1e-9: return None 
        X = (math.cos(dec_rad) * math.sin(ra_rad - ra0_rad)) / cos_c
        Y = (math.cos(dec0_rad) * math.sin(dec_rad) - 
             math.sin(dec0_rad) * math.cos(dec_rad) * math.cos(ra_rad - ra0_rad)) / cos_c
        return X, Y

    def calculate_and_set_pole_position(self, silent=False):
        if len(self.polar_stars_data) < 3:
            if not silent: self.status_var.set("Нужно выбрать 3 звезды для расчета полюса."); 
            return False 
        src_pts_sky_proj = []; dst_pts_img = []      
        polaris_sky_coords = self.polar_star_sky_coords_deg['polaris']
        ra0_deg, dec0_deg = polaris_sky_coords['ra'], polaris_sky_coords['dec']
        for star_key in self.polar_star_names_ordered:
            entry = self.polar_stars_data.get(star_key)
            if not entry or 'screen_coords' not in entry or 'sky_coords' not in entry:
                if not silent: self.status_var.set(f"Ошибка: данные для {star_key} отсутствуют."); return False
            img_coords = entry['screen_coords']; sky_coords_celestial = entry['sky_coords']
            sky_proj_coords = self.gnomonic_projection(sky_coords_celestial['ra'], sky_coords_celestial['dec'], ra0_deg, dec0_deg)
            if sky_proj_coords is None:
                if not silent: self.status_var.set(f"Ошибка проекции для {star_key}."); return False
            src_pts_sky_proj.append(list(sky_proj_coords)); dst_pts_img.append(list(img_coords))
        src_pts_np = np.float32(src_pts_sky_proj); dst_pts_np = np.float32(dst_pts_img)
        try:
            self.affine_transform_matrix = cv2.getAffineTransform(src_pts_np, dst_pts_np)
            if self.affine_transform_matrix is None: raise ValueError("getAffineTransform вернул None")
            ncp_sky_proj = self.gnomonic_projection(ra0_deg, 90.0, ra0_deg, dec0_deg) 
            if ncp_sky_proj is None: ncp_sky_proj = (0.0, 0.0) 
            ncp_sky_proj_homogeneous = np.array([[ncp_sky_proj[0]], [ncp_sky_proj[1]], [1.0]], dtype=np.float32)
            pole_img_coords_transformed = self.affine_transform_matrix @ ncp_sky_proj_homogeneous
            self.polar_pole_calculated_coords = (int(pole_img_coords_transformed[0,0]), int(pole_img_coords_transformed[1,0]))
            if not silent: self.status_var.set(f"Полюс (аффин.): {self.polar_pole_calculated_coords}. Можно показать сетку.")
            print(f"Полюс (аффин.): {self.polar_pole_calculated_coords}")
            p_img=self.polar_stars_data['polaris']['screen_coords']; h1_img=self.polar_stars_data['hip7283']['screen_coords']
            p_sky_cel=self.polar_stars_data['polaris']['sky_coords']; h1_sky_cel=self.polar_stars_data['hip7283']['sky_coords']
            img_dist_px=math.sqrt((p_img[0]-h1_img[0])**2+(p_img[1]-h1_img[1])**2)
            delta_ra_rad=math.radians(p_sky_cel['ra']-h1_sky_cel['ra'])
            dec_p_rad=math.radians(p_sky_cel['dec']); dec_h1_rad=math.radians(h1_sky_cel['dec'])
            val_acos=math.sin(dec_p_rad)*math.sin(dec_h1_rad)+math.cos(dec_p_rad)*math.cos(dec_h1_rad)*math.cos(delta_ra_rad)
            val_acos=max(-1.0,min(1.0,val_acos)); angular_dist_deg=math.degrees(math.acos(val_acos))
            if angular_dist_deg > 1e-3: self.pixels_per_degree_estimate = img_dist_px / angular_dist_deg
            print(f"Масштаб: {self.pixels_per_degree_estimate:.2f} px/deg (Polaris-HIP7283: угл.расст {angular_dist_deg:.3f} deg)")
            return True
        except Exception as e:
            traceback.print_exc()
            if not silent: self.status_var.set(f"Ошибка расчета полюса: {e}")
            print(f"Ошибка расчета полюса (аффинное преобразование): {e}")
            self.polar_pole_calculated_coords = None; self.affine_transform_matrix = None; return False

    def toggle_polar_grid_display(self):
        if not self.polar_pole_calculated_coords: messagebox.showwarning("Сетка", "Сначала рассчитайте полюс."); return
        self.show_polar_grid = not self.show_polar_grid
        self.toggle_grid_btn.config(text="Скрыть сетку" if self.show_polar_grid else "Показать сетку")
        self.status_var.set(f"Полюс: {self.polar_pole_calculated_coords}. Сетка {'вкл.' if self.show_polar_grid else 'выкл.'}")

    def draw_polar_grid_on_frame(self, frame_bgr, pole_coords_on_screen):
        pole_x, pole_y = int(pole_coords_on_screen[0]), int(pole_coords_on_screen[1])
        grid_color_bgr = (0, 0, 255) # Красный в BGR
        
        # Рисуем полюс с помощью cv2
        cv2.circle(frame_bgr, (pole_x, pole_y), 5, (0, 255, 255), -1) # Желтый центр (BGR)
        cv2.circle(frame_bgr, (pole_x, pole_y), 8, grid_color_bgr, 1) # Красный круг
        
        # Текст "NCP" рисуем через PIL, передавая BGR цвет
        frame_bgr = self.draw_text_pil(frame_bgr, "NCP", (pole_x + 10, pole_y - 20), self.pil_font_overlay, grid_color_bgr)
        
        max_r_draw = int(min(frame_bgr.shape[0], frame_bgr.shape[1]) * 0.6) 
        for dec_offset_deg in [0.5, 1, 1.5, 2, 2.5, 3]: 
            radius_px = int(dec_offset_deg * self.pixels_per_degree_estimate)
            if radius_px > max_r_draw : continue 
            if radius_px > 5: 
                cv2.circle(frame_bgr, (pole_x, pole_y), radius_px, grid_color_bgr, 1)
                frame_bgr = self.draw_text_pil(frame_bgr, f"{dec_offset_deg}deg", 
                                               (pole_x + radius_px + 5, pole_y + 5), 
                                               self.pil_font_overlay, grid_color_bgr)
        for angle_deg_sky in range(0, 360, 30):
            angle_rad_draw = math.radians(angle_deg_sky) 
            end_x = int(pole_x + max_r_draw * math.cos(angle_rad_draw)); end_y = int(pole_y + max_r_draw * math.sin(angle_rad_draw))
            cv2.line(frame_bgr, (pole_x, pole_y), (end_x, end_y), grid_color_bgr, 1)
        return frame_bgr 


    def toggle_center_crosshair(self): 
        self.show_center_crosshair = not self.show_center_crosshair
        status_text = "Центр. перекрестие: ВКЛ" if self.show_center_crosshair else "Центр. перекрестие: ВЫКЛ"
        print(status_text)
        current_status = self.status_var.get()
        if not self.polar_align_mode or "Кликните на" not in current_status :
             self.status_var.set(status_text)


    def on_closing(self):
        print("Закрытие приложения...")
        self.is_capturing = False 
        if self.capture_thread and self.capture_thread.is_alive(): print("Ожидание потока..."); self.capture_thread.join(timeout=1.0) 
        if self.camera:
            try: print("Остановка/закрытие камеры..."); self.camera.stop_video_capture(); self.camera.close(); self.camera = None; print("Камера закрыта.")
            except Exception as e: print(f"Ошибка закрытия камеры: {e}")
        if hasattr(self, 'root') and self.root.winfo_exists(): self.root.destroy()
        print("Приложение закрыто.")

    def run(self):
        self.root.mainloop()

if __name__ == "__main__":
    app = CameraTrackerApp()
    app.run()
